"""
Configuration parsing for defining problem review requirements.

The configuration file is also used for manually tracking the review status
of the different parts of each problem.
"""

import itertools
import logging
import os
import tomllib
from dataclasses import dataclass
from typing import Any

from crifx.contest_objects import LanguageGroup, ProgrammingLanguage

CONFIG_FILENAME = "crifx.toml"


def _get_or_default(key: str, dct: dict[str, Any], obj: Any) -> Any:
    """Get a value from a dictionary or from a fallback object."""
    if key in dct:
        return dct.get(key)
    return getattr(obj, key)


@dataclass(frozen=True)
class ReviewCountRequirements:
    """Requirements applicable to each problem."""

    # The number of distinct judges required to have provided an AC submission.
    independent_ac: int
    # The number of language groups required to have an AC submission.
    language_groups_ac: int
    # The number of wrong_answer submissions required.
    submissions_wa: int
    # The number of tle submissions required.
    submissions_tle: int
    # The number of judges required to have reviewed the problem statement.
    statement_reviewers: int
    # The number of judges required to have reviewed the input validator(s).
    validator_reviewers: int
    # The number of judges required to have reviewed the test data.
    data_reviewers: int

    @staticmethod
    def from_toml_dict(toml_dict: dict[str, int]) -> "ReviewCountRequirements":
        """Initialize ReviewCountRequirements from a toml dict."""
        return ReviewCountRequirements(
            _get_or_default("independent_ac", toml_dict, REVIEW_COUNT_DEFAULT),
            _get_or_default("language_groups_ac", toml_dict, REVIEW_COUNT_DEFAULT),
            _get_or_default("submissions_wa", toml_dict, REVIEW_COUNT_DEFAULT),
            _get_or_default("submissions_tle", toml_dict, REVIEW_COUNT_DEFAULT),
            _get_or_default("statement_reviewers", toml_dict, REVIEW_COUNT_DEFAULT),
            _get_or_default("validator_reviewers", toml_dict, REVIEW_COUNT_DEFAULT),
            _get_or_default("data_reviewers", toml_dict, REVIEW_COUNT_DEFAULT),
        )


REVIEW_COUNT_DEFAULT = ReviewCountRequirements(
    independent_ac=3,
    language_groups_ac=2,
    submissions_wa=1,
    submissions_tle=1,
    statement_reviewers=3,
    validator_reviewers=2,
    data_reviewers=2,
)


@dataclass(frozen=True)
class LanguageGroupConfig:
    """Configuration for requirements for a language group."""

    # Identifier name used for the language group.
    identifier: str
    # The programming languages in the group.
    language_group: LanguageGroup
    # The number of required AC submissions from this group.
    required_ac_count: int = 0

    @staticmethod
    def from_toml_dict(toml_dict: dict[str, Any]) -> "LanguageGroupConfig":
        """Initialize a LanguageGroupConfig from a toml dict."""
        group_identifier = toml_dict.get("name")
        if group_identifier is None:
            raise ValueError(
                "Language group in the `crifx.toml` file is missing a 'name'"
            )
        language_names = toml_dict.get("languages", [])
        languages = []
        for language_name in language_names:
            language = ProgrammingLanguage.from_language_name(language_name)
            if language is None:
                logging.warning(
                    "Language %s is not recognized by crifx.", language_name
                )
            else:
                languages.append(language)
        return LanguageGroupConfig(
            identifier=group_identifier,
            language_group=LanguageGroup(*languages),
            required_ac_count=toml_dict.get("required_ac_count", 0),
        )


@dataclass(frozen=True)
class AliasGroup:
    """Alias group object for parsing alternate names from a config file."""

    identifier: str
    git_name: str | None
    aliases: list[str]

    @staticmethod
    def from_toml_dict(toml_dict: dict[str, Any]) -> "AliasGroup":
        """Parse an AliasGroup from a toml dictionary."""
        primary_name = toml_dict.get("primary_name")
        if primary_name is None:
            raise ValueError(
                "The `crifx.toml` file is missing a 'primary_name' for one or more 'judge' tables."
            )
        git_name = toml_dict.get("git_name")
        aliases = toml_dict.get("aliases", [])
        return AliasGroup(primary_name, git_name, aliases)


class Config:
    """Configuration for crifx requirements and review status."""

    def __init__(self, toml_dict):
        self.github_repo_url = toml_dict.get("github_repo_url")
        self.review_requirements = ReviewCountRequirements.from_toml_dict(
            toml_dict.get("review_requirements", {}),
        )
        self.language_group_configs = []
        self.alias_groups = []
        language_groups = toml_dict.get("language_group", [])
        for language_group_dict in language_groups:
            language_group_config = LanguageGroupConfig.from_toml_dict(
                language_group_dict,
            )
            if not language_group_config.language_group.languages:
                # No languages parsed from the language group.
                continue
            self.language_group_configs.append(language_group_config)
        alias_groups = toml_dict.get("judge", [])
        for alias_group_dict in alias_groups:
            alias_group = AliasGroup.from_toml_dict(alias_group_dict)
            self.alias_groups.append(alias_group)
        for alias_group_1, alias_group_2 in itertools.product(
            self.alias_groups, self.alias_groups
        ):
            if alias_group_1 == alias_group_2:
                continue
            shared_aliases = list(
                set(alias_group_1.aliases) & set(alias_group_2.aliases)
            )
            if shared_aliases:
                logging.warning(
                    "Judges '%s' and '%s' share the following aliases: %s. Remove shared "
                    "aliases from the crifx configuration file.",
                    alias_group_1.identifier,
                    alias_group_2.identifier,
                )

    @property
    def track_review_status(self) -> bool:
        """Return True iff manual reviews are required to be tracked in files per problem."""
        return (
            self.review_requirements.statement_reviewers > 0
            or self.review_requirements.data_reviewers > 0
            or self.review_requirements.validator_reviewers > 0
        )


def parse_config(problemset_root_path: str) -> Config:
    """Parse a configuration file into a Config object."""
    config_path = os.path.join(problemset_root_path, CONFIG_FILENAME)
    with open(config_path, "rb") as config_file:
        toml_dict = tomllib.load(config_file)
    return Config(toml_dict)



"""Pytest fixtures."""

import os
import uuid
import zipfile
from pathlib import Path

import pygit2
import pytest

from crifx.contest_objects import Judge, Judgement, ProgrammingLanguage, Submission
from crifx.git_manager import GitUser


@pytest.fixture(scope="session", autouse=True)
def remove_git_config_paths():
    """Remove git configurations from pygit2 to improve test reproducibility."""
    levels = [
        pygit2.enums.ConfigLevel.GLOBAL,
        pygit2.enums.ConfigLevel.XDG,
        pygit2.enums.ConfigLevel.SYSTEM,
    ]
    for level in levels:
        pygit2.settings.search_path[level] = ""


class TemporaryRepository:
    """
    Context managed class for creating a repository at a temporary path.

    The provided name must be a .zip file name.
    """

    def __init__(self, name, tmp_path):
        self.name = name
        self.tmp_path = tmp_path

    def __enter__(self):
        path = Path(__file__).parent.joinpath("data", self.name)
        temp_repo_path = Path(self.tmp_path).joinpath(path.stem)
        with zipfile.ZipFile(path) as zipped_data_file:
            zipped_data_file.extractall(self.tmp_path)
        return temp_repo_path

    def __exit__(self, exc_type, exc_value, traceback):
        pass


@pytest.fixture
def empty_repo(tmp_path):
    """Initialise a temporary empty (but non-bare) repository."""
    with TemporaryRepository("empty_repo.zip", tmp_path) as path:
        yield pygit2.Repository(path)


@pytest.fixture
def global_git_config_path(tmp_path):
    """Set and return a global git config path."""
    level = pygit2.enums.ConfigLevel.GLOBAL
    pygit2.settings.search_path[level] = str(tmp_path)
    yield os.path.join(tmp_path, ".gitconfig")


@pytest.fixture
def tmp_file_path(tmp_path):
    """Get a unique file path in a tmp directory."""
    unique_id = str(uuid.uuid4())
    tmp_file = os.path.join(tmp_path, f"crifx-{unique_id}.txt")
    yield tmp_file
    os.remove(tmp_file)


@pytest.fixture
def make_file_at_path():
    """Write a file in the provided directory and delete it afterwards."""
    written_files = []

    def _func(dir_path):
        unique_id = str(uuid.uuid4())
        file_name = f"crifx-tmp_file-{unique_id}.txt"
        file_path = os.path.join(dir_path, file_name)
        open(file_path, "a").close()
        written_files.append(file_path)
        return file_path

    yield _func
    for file in written_files:
        os.remove(file)


@pytest.fixture
def examples_path(pytestconfig):
    """Get the path to the examples directory."""
    yield os.path.join(pytestconfig.rootpath, "examples")


@pytest.fixture
def scenarios_path(pytestconfig):
    """Get the path to the scenarios directory."""
    yield os.path.join(pytestconfig.rootpath, "tests", "scenarios")


@pytest.fixture
def make_problem_skeleton_dir(tmp_path):
    """Make a problem directory skeleton."""

    def _func() -> str:
        problem_name = f"problem-{str(uuid.uuid4())}"
        problem_path = os.path.join(tmp_path, problem_name)
        os.mkdir(problem_path)
        os.makedirs(os.path.join(problem_path, "data/sample"))
        os.makedirs(os.path.join(problem_path, "data/secret"))
        os.makedirs(os.path.join(problem_path, "submissions/accepted"))
        os.mkdir(os.path.join(problem_path, "problem_statement"))
        os.mkdir(os.path.join(problem_path, "input_format_validators"))
        return problem_path

    yield _func


@pytest.fixture
def make_judged_lang_submission():
    """Make a submission with a given judgement and programming language."""

    def _func(judgement: Judgement, language: ProgrammingLanguage) -> Submission:
        return Submission(Judge("name", None), "solution", language, judgement, 1, 1)

    yield _func


@pytest.fixture
def make_authored_submission():
    """Make a submission with a Judge with the given primary name and git name."""

    def _func(
        primary_name: str,
        git_name: str | None,
        judgement: Judgement = Judgement.ACCEPTED,
        language: ProgrammingLanguage = ProgrammingLanguage.JAVA,
    ) -> Submission:
        if git_name is None:
            judge = Judge(primary_name, None)
        else:
            git_email = f"{git_name}@example.com"
            git_user = GitUser(
                git_name, git_email, str.encode(git_name), str.encode(git_email)
            )
            judge = Judge(primary_name, git_user)
        return Submission(judge, "solution", language, judgement, 0, 0)

    yield _func



"""Logic for parsing a ProblemSet object from a git directory."""

import logging
import os
import re
import tomllib
from typing import Any, Optional

from crifx.config_parser import AliasGroup
from crifx.contest_objects import (
    UNKNOWN_JUDGE,
    Judge,
    Judgement,
    Problem,
    ProblemSet,
    ProblemTestCase,
    ProgrammingLanguage,
    Submission,
)
from crifx.dir_layout_parsing import get_problem_root_dirs, is_contest_problems_root
from crifx.git_manager import GitManager, GitUser
from crifx.report_objects import (
    DEFAULT_REVIEW_STATUS,
    DEFAULT_REVIEW_STATUS_TOML,
    ReviewStatus,
)

TEST_CASE_IMAGE_EXTENSIONS = ["png", "jpg", "jpeg"]
PROBLEM_REVIEW_STATUS_FILENAME = "crifx-problem-status.toml"
CRIFX_AUTHOR_PATTERN = "crifx!\(author=([a-zA-Z0-9_ ]+)\)"


class ProblemSetParser:
    """Controller for parsing a ProblemSet object."""

    def __init__(
        self,
        problemset_root_path: str,
        git_manager: GitManager,
        alias_groups: list[AliasGroup],
        track_review_status: bool,
    ):
        if not is_contest_problems_root(problemset_root_path):
            raise ValueError(
                f"Path '{problemset_root_path}' is not a problemset root path."
            )
        self.problemset_root_path = problemset_root_path
        self.git_manager = git_manager
        self.track_review_status = track_review_status
        self.judges_by_name: dict[str, Judge] = {}
        self._set_judges_by_name(alias_groups)

    def _set_judges_by_name(self, alias_groups: list[AliasGroup]):
        git_users = self.git_manager.get_committers_and_authors()
        git_users_by_name: dict[str, GitUser] = {}
        for user in git_users:
            if user.name in git_users:
                logging.warning(
                    "Multiple users with name %s: (%s), (%s). "
                    "Commits from these users may be reported as from the same person.",
                    user.name,
                    git_users_by_name[user.name],
                    user,
                )
            git_users_by_name[user.name] = user
        gitless = []
        aliases_by_git_name = {}
        for alias_group in alias_groups:
            if alias_group.git_name is not None:
                aliases_by_git_name[alias_group.git_name] = alias_group.aliases
            else:
                gitless.append(alias_group)
        self.judges_by_name = {
            git_user.name: Judge(
                git_user.name, git_user, *aliases_by_git_name.get(git_user.name, [])
            )
            for git_user in git_users_by_name.values()
        }
        for alias_group in alias_groups:
            if (
                alias_group.git_name is not None
                and alias_group.git_name not in self.judges_by_name
            ):
                # The user has a git name but no commits in the repository yet.
                self.judges_by_name[alias_group.git_name] = Judge(
                    alias_group.identifier,
                    GitUser(alias_group.git_name, "unknown email", b"", b""),
                    *alias_group.aliases,
                )
        for alias_group in gitless:
            if alias_group.identifier in git_users_by_name:
                logging.warning(
                    "Add a 'git_name' for judge '%s' in the crifx configuration "
                    "file to use their aliases.",
                    alias_group.identifier,
                )
            else:
                self.judges_by_name[alias_group.identifier] = Judge(
                    alias_group.identifier, None, *alias_group.aliases
                )
        logging.debug("Identified judges: %s", str(self.judges_by_name))

    def parse_problemset(self) -> ProblemSet:
        """Parse a ProblemSet."""
        problem_root_dirs = get_problem_root_dirs(self.problemset_root_path)
        problems = []
        for problem_root_dir in problem_root_dirs:
            problem = self._parse_problem(problem_root_dir)
            problems.append(problem)
        return ProblemSet(problems)

    def _parse_problem(self, problem_root_dir: str) -> Problem:
        """Parse a problem object from a problem directory."""
        _, name = os.path.split(problem_root_dir)
        problem_test_cases = self._parse_problem_test_cases(problem_root_dir)
        problem_test_cases.sort(key=lambda x: x.sort_key())
        submissions = self._parse_submissions(problem_root_dir)
        review_status = self._parse_review_status(problem_root_dir)
        return Problem(name, problem_test_cases, submissions, review_status)

    def _parse_problem_test_cases(self, problem_root_dir: str) -> list[ProblemTestCase]:
        """Parse the problem test cases from a problem directory."""
        sample_data_dir = os.path.join(problem_root_dir, "data", "sample")
        secret_data_dir = os.path.join(problem_root_dir, "data", "secret")
        return self._parse_test_case_dir(sample_data_dir) + self._parse_test_case_dir(
            secret_data_dir
        )

    def _parse_test_case_dir(self, test_case_dir: str):
        """Parse the test cases from a test case directory."""
        if not os.path.exists(test_case_dir):
            return []
        in_files = set()
        ans_files = set()
        test_cases = []
        for filename in os.listdir(test_case_dir):
            file_path = os.path.join(test_case_dir, filename)
            if os.path.isfile(file_path):
                if filename.endswith(".in"):
                    in_files.add(filename[:-3])
                elif filename.endswith(".ans"):
                    ans_files.add(filename[:-4])
            elif os.path.isdir(file_path):
                nested_dir = os.path.join(test_case_dir, filename)
                test_cases.extend(self._parse_test_case_dir(nested_dir))
        for in_filename in in_files:
            if in_filename not in ans_files:
                logging.warning(
                    "Input file %s.in has no corresponding .ans file.",
                    os.path.join(test_case_dir, in_filename),
                )
        for ans_filename in ans_files:
            if ans_filename not in in_files:
                logging.warning(
                    "Answer file %s.ans has no corresponding .in file.",
                    os.path.join(test_case_dir, ans_filename),
                )
        for filename in in_files:
            if filename not in ans_files:
                continue
            name = filename
            is_sample = test_case_dir.startswith("sample")
            desc_lines = []
            image_extension = None
            try:
                desc_file_path = os.path.join(test_case_dir, f"{filename}.desc")
                image_extension = None
                for extension in TEST_CASE_IMAGE_EXTENSIONS:
                    if os.path.exists(
                        os.path.join(test_case_dir, f"{filename}.{extension}")
                    ):
                        image_extension = extension
                        break
                if os.path.exists(desc_file_path):
                    with open(desc_file_path) as desc_file:
                        desc_lines = list(desc_file.readlines())
            except (FileExistsError, FileNotFoundError, PermissionError):
                logging.exception("Test case file could not be read.")
            test_case = ProblemTestCase(
                name,
                is_sample,
                test_case_dir,
                desc_lines,
                image_extension,
            )
            test_cases.append(test_case)
        return test_cases

    def _parse_submissions(self, problem_root_dir: str) -> list[Submission]:
        """Parse the submissions from a problem directory."""
        ac_dir = os.path.join(problem_root_dir, "submissions", "accepted")
        wa_dir = os.path.join(problem_root_dir, "submissions", "wrong_answer")
        tle_dir = os.path.join(problem_root_dir, "submissions", "time_limit_exceeded")
        rte_dir = os.path.join(problem_root_dir, "submissions", "run_time_error")
        submissions = []
        submissions.extend(self._parse_submissions_dir(ac_dir, Judgement.ACCEPTED))
        submissions.extend(self._parse_submissions_dir(wa_dir, Judgement.WRONG_ANSWER))
        submissions.extend(
            self._parse_submissions_dir(tle_dir, Judgement.TIME_LIMIT_EXCEEDED)
        )
        submissions.extend(
            self._parse_submissions_dir(rte_dir, Judgement.RUN_TIME_ERROR)
        )
        return submissions

    def _parse_submissions_dir(
        self, submissions_dir: str, judgement: Judgement
    ) -> list[Submission]:
        """Parse the Submission objects from a directory."""
        if not os.path.exists(submissions_dir):
            return []
        submissions = []
        for filename in os.listdir(submissions_dir):
            language = ProgrammingLanguage.from_filename(filename)
            if language is None:
                continue
            submission_path = os.path.join(submissions_dir, filename)
            lines_of_code = 0
            file_bytes = 0
            author_name_override = None
            try:
                with open(submission_path, "r") as submission_file:
                    submission_lines = submission_file.readlines()
                    for line_number, line in enumerate(submission_lines):
                        author_match = re.search(CRIFX_AUTHOR_PATTERN, line)
                        if author_match is not None:
                            author_name_override = author_match.group(1)
                            logging.debug(
                                "Found author override for file %s on line %d. Override name is '%s'",
                                submission_path,
                                line_number + 1,
                                author_name_override,
                            )
                            break
                    lines_of_code = len(submission_lines)
                file_bytes = os.stat(submission_path).st_size
            except (FileExistsError, FileNotFoundError, PermissionError):
                logging.warning(
                    "Could not determine size of submission at path '%s'",
                    submission_path,
                )
            git_user_guess = self.git_manager.guess_file_author(submission_path)
            filename_guess = self.guess_author_by_filename(filename)
            if author_name_override is not None:
                judge = UNKNOWN_JUDGE
                for candidate_judge in self.judges_by_name.values():
                    if candidate_judge.has_alias(author_name_override):
                        judge = candidate_judge
                        break
            elif filename_guess is not None:
                judge = filename_guess

            else:
                judge = (
                    self.judges_by_name.get(getattr(git_user_guess, "name"))
                    or UNKNOWN_JUDGE
                )
            submission = Submission(
                judge,
                filename,
                language,
                judgement,
                lines_of_code,
                file_bytes,
            )
            submissions.append(submission)
        return submissions

    def _parse_review_status(
        self,
        problem_root_dir: str,
    ) -> ReviewStatus:
        """Parse the problem review status."""
        if not self.track_review_status:
            return DEFAULT_REVIEW_STATUS
        review_status_path = os.path.join(
            problem_root_dir,
            PROBLEM_REVIEW_STATUS_FILENAME,
        )
        if not os.path.exists(review_status_path):
            with open(review_status_path, "w") as review_status_file:
                review_status_file.write(DEFAULT_REVIEW_STATUS_TOML)
        try:
            with open(review_status_path, "rb") as review_status_file:
                toml_dict = tomllib.load(review_status_file)
        except (PermissionError, FileNotFoundError, FileExistsError):
            logging.exception(
                "Failed to read problem review status file at path '%s'.",
                review_status_path,
            )
            return DEFAULT_REVIEW_STATUS
        github_issue_id = toml_dict.get("github_issue_id")
        if not isinstance(github_issue_id, int):
            github_issue_id = None
        review_status_dict = toml_dict.get("review_status", {})
        statement_reviewed_by = _read_reviewers(
            review_status_dict, "statement_reviewed_by", review_status_path
        )
        validators_reviewed_by = _read_reviewers(
            review_status_dict, "validators_reviewed_by", review_status_path
        )
        data_reviewed_by = _read_reviewers(
            review_status_dict, "data_reviewed_by", review_status_path
        )
        return ReviewStatus(
            github_issue_id,
            statement_reviewed_by,
            validators_reviewed_by,
            data_reviewed_by,
        )

    def guess_author_by_filename(self, filename) -> Judge | None:
        """Guess the author of a file based on the filename and configured aliases."""
        if "." in filename:
            filename_without_extension = filename[: filename.rindex(".")]
        else:
            filename_without_extension = filename
        underscore_split_filename = filename_without_extension.split("_")
        for part in underscore_split_filename:
            for judge in self.judges_by_name.values():
                if judge.has_alias(part):
                    return judge
        return None


def _read_reviewers(
    review_status_dict: dict[str, Any], reviewer_str: str, path: str
) -> list[str]:
    """Get a list of reviewers strings from a toml dictionary."""
    reviewers = review_status_dict.get(reviewer_str, [])
    if not isinstance(reviewers, list) or not all(
        isinstance(val, str) for val in reviewers
    ):
        logging.error(
            "%s [review_status] %s should be a list of strings, but instead it is %s",
            path,
            reviewer_str,
            reviewers,
        )
        reviewers = []
    return reviewers



"""Tests for the ProgrammingLanguage enumeration object."""

from crifx.contest_objects import ProgrammingLanguage


def test_guess_language():
    """Each recognized programming language is identified correctly."""
    cases = [
        ("sol.c", ProgrammingLanguage.C),
        ("sol.cc", ProgrammingLanguage.CPP),
        ("sol.cpp", ProgrammingLanguage.CPP),
        ("sol.py", ProgrammingLanguage.PYTHON),
        ("sol.java", ProgrammingLanguage.JAVA),
        ("sol.kt", ProgrammingLanguage.KOTLIN),
        ("sol.rs", ProgrammingLanguage.RUST),
        ("validator.ctd", ProgrammingLanguage.CTD),
        ("validator.viva", ProgrammingLanguage.VIVA),
        ("sol.bf", None),
        ("sol.extra_period.py", ProgrammingLanguage.PYTHON),
        ("path/to/file.java", ProgrammingLanguage.JAVA),
    ]
    # Languages are guessed from filenames as expected.
    for filename, expected in cases:
        assert ProgrammingLanguage.from_filename(filename) is expected
    # All languages are tested.
    languages_tested = set(language for _, language in cases if language is not None)
    assert len(languages_tested) == len(ProgrammingLanguage)



"""Data structure for a problem submission."""

from dataclasses import dataclass

from crifx.contest_objects.judge import Judge
from crifx.contest_objects.judgement import Judgement
from crifx.contest_objects.programming_language import ProgrammingLanguage


@dataclass
class Submission:
    """Data structure for a problem submission."""

    author: Judge
    filename: str
    language: ProgrammingLanguage
    judgement: Judgement
    lines_of_code: int
    bytes_count: int



"""Tests for methods on the Problem object."""

from crifx.contest_objects import Judgement, Problem, ProgrammingLanguage
from crifx.report_objects import DEFAULT_REVIEW_STATUS


def test_language_count(make_judged_lang_submission):
    """Test counting the number of submissions for each language."""
    ac_java_1 = make_judged_lang_submission(
        Judgement.ACCEPTED,
        ProgrammingLanguage.JAVA,
    )
    ac_java_2 = make_judged_lang_submission(
        Judgement.ACCEPTED, ProgrammingLanguage.JAVA
    )
    ac_c = make_judged_lang_submission(Judgement.ACCEPTED, ProgrammingLanguage.C)
    wa_java = make_judged_lang_submission(
        Judgement.WRONG_ANSWER, ProgrammingLanguage.JAVA
    )
    tle_kotlin = make_judged_lang_submission(
        Judgement.TIME_LIMIT_EXCEEDED, ProgrammingLanguage.KOTLIN
    )
    rte_rust = make_judged_lang_submission(
        Judgement.RUN_TIME_ERROR, ProgrammingLanguage.RUST
    )
    problem = Problem(
        "problem",
        [],
        [ac_java_1, tle_kotlin, wa_java, rte_rust, ac_c, ac_java_2],
        DEFAULT_REVIEW_STATUS,
    )
    ac_languages = problem.ac_languages()
    assert ac_languages == {ProgrammingLanguage.JAVA: 2, ProgrammingLanguage.C: 1}
    assert sorted(problem.ac_submissions, key=lambda x: x.language.value) == [
        ac_c,
        ac_java_1,
        ac_java_2,
    ]
    assert problem.wa_submissions == [wa_java]
    assert problem.tle_submissions == [tle_kotlin]
    assert problem.rte_submissions == [rte_rust]
    problem = Problem("problem", [], [], DEFAULT_REVIEW_STATUS)
    ac_languages = problem.ac_languages()
    assert ac_languages == {}
    assert problem.ac_submissions == []
    assert problem.wa_submissions == []
    assert problem.tle_submissions == []
    assert problem.rte_submissions == []


def test_independent_author_count(make_authored_submission):
    """Test counting the number of submissions from distinct judges."""
    finn_java_ac_1 = make_authored_submission(
        "Finn", None, Judgement.ACCEPTED, ProgrammingLanguage.JAVA
    )
    finn_java_ac_2 = make_authored_submission(
        "Finn", None, Judgement.ACCEPTED, ProgrammingLanguage.JAVA
    )
    finn_rust_ac = make_authored_submission(
        "Finn", None, Judgement.ACCEPTED, ProgrammingLanguage.RUST
    )
    alice_c_ac = make_authored_submission(
        "Alice", "alice", Judgement.ACCEPTED, ProgrammingLanguage.C
    )
    alice_c_wa = make_authored_submission(
        "Alice", "alice", Judgement.WRONG_ANSWER, ProgrammingLanguage.C
    )
    bob_java_tle = make_authored_submission(
        "Bob", "bob", Judgement.TIME_LIMIT_EXCEEDED, ProgrammingLanguage.JAVA
    )
    problem = Problem(
        "problem", [], [finn_java_ac_1, alice_c_ac, bob_java_tle], DEFAULT_REVIEW_STATUS
    )
    assert problem.independent_ac_count() == 2
    problem = Problem(
        "problem", [], [finn_java_ac_1, finn_rust_ac], DEFAULT_REVIEW_STATUS
    )
    assert problem.independent_ac_count() == 1
    problem = Problem("problem", [], [alice_c_wa, bob_java_tle], DEFAULT_REVIEW_STATUS)
    assert problem.independent_ac_count() == 0
    problem = Problem(
        "problem",
        [],
        [finn_java_ac_1, finn_java_ac_2, bob_java_tle],
        DEFAULT_REVIEW_STATUS,
    )
    assert problem.independent_ac_count() == 1



#!/bin/python3

"""Solution to the Hello World problem."""

# Use a crifx command to override the author for this submission instead
# of relying on git blame information.
# crifx!(author=Homer Simpson)

from sys import stdin

line = stdin.readline().strip()
print(f"Hello {line}")



#!/bin/python3

"""Loop forever, then print the first line of input."""

from sys import stdin

val = 0
while True:
    val = 1 - val
    if val == 2:
        break

line = stdin.readline()
print(line)



"""Programming language enumeration object."""

from enum import Enum
from typing import Union


class ProgrammingLanguage(Enum):
    """Recognised programming languages."""

    C = "C"
    CPP = "C++"
    PYTHON = "Python"
    JAVA = "Java"
    KOTLIN = "Kotlin"
    RUST = "Rust"
    CTD = "Checktestdata"
    VIVA = "Viva"

    def file_extensions(self) -> list[str]:
        """Get a list of recognised file extensions for the language."""
        match self:
            case ProgrammingLanguage.C:
                return ["c"]
            case ProgrammingLanguage.CPP:
                return ["cc", "cpp", "cxx", "C"]
            case ProgrammingLanguage.PYTHON:
                return ["py"]
            case ProgrammingLanguage.JAVA:
                return ["java"]
            case ProgrammingLanguage.KOTLIN:
                return ["kt"]
            case ProgrammingLanguage.RUST:
                return ["rs"]
            case ProgrammingLanguage.CTD:
                return ["ctd"]
            case ProgrammingLanguage.VIVA:
                return ["viva"]

    @staticmethod
    def from_filename(filename: str) -> Union["ProgrammingLanguage", None]:
        """Guess the programming language from a file name."""
        if "." not in filename:
            return None
        extension = filename[filename.rindex(".") + 1 :]
        for programming_language_str in ProgrammingLanguage:
            programming_language = ProgrammingLanguage(programming_language_str)
            if extension in programming_language.file_extensions():
                return programming_language
        return None

    @staticmethod
    def from_language_name(language_name: str) -> Union["ProgrammingLanguage", None]:
        """Get the programming language from a language name string."""
        for programming_language in ProgrammingLanguage:
            if language_name.lower() == programming_language.value.lower():
                return programming_language
        return None


class LanguageGroup:
    """A group of programming languages."""

    def __init__(self, *languages: ProgrammingLanguage):
        self.languages = tuple(sorted(languages, key=lambda x: x.value))

    def has_language(self, language: ProgrammingLanguage) -> bool:
        """Return True if the given language is in the group."""
        return language in self.languages

    def __eq__(self, other):
        if not isinstance(other, LanguageGroup):
            return False
        return self.languages == other.languages

    def __hash__(self):
        return hash(self.languages)



"""Problem test case."""

import os
from dataclasses import dataclass


@dataclass
class ProblemTestCase:
    """Problem test case data."""

    name: str
    is_sample: bool
    dir_path: str
    description_lines: list[str]
    image_extension: str | None

    @property
    def has_description(self):
        """Return true if there is a nonempty description for the test case."""
        return bool(self.description_lines)

    def _get_path(self, extension) -> str:
        return os.path.join(self.dir_path, f"{self.name}.{extension}")

    @property
    def input_path(self) -> str:
        """Get the path to the input file."""
        return self._get_path("in")

    @property
    def answer_path(self) -> str:
        """Get the path to the answer file."""
        return self._get_path("ans")

    @property
    def description_path(self) -> str:
        """Get the path to the description file."""
        return self._get_path("desc")

    @property
    def image_path(self) -> str | None:
        """Get the path to the image file."""
        if self.image_extension is None:
            return None
        return self._get_path(self.image_extension)

    def sort_key(self) -> tuple[bool, str, str]:
        """Get the sort key for this test case."""
        return not self.is_sample, self.dir_path, self.name



#!/bin/python3

"""
This file uses 'jd' in the filename.

This is an alias defined for Jane Doe in the crifx.toml file.
"""

from sys import stdin

tokens = stdin.readline().split(" ")
print(int(tokens[0]) + int(tokens[1]))



"""Module for data tracking objects used by crifx."""

from crifx.report_objects.review_status import (
    DEFAULT_REVIEW_STATUS,
    DEFAULT_REVIEW_STATUS_TOML,
    ReviewStatus,
)

__all__ = [
    "DEFAULT_REVIEW_STATUS",
    "DEFAULT_REVIEW_STATUS_TOML",
    "ReviewStatus",
]



"""ReviewStatus object for tracking information about a problem."""

from dataclasses import dataclass


@dataclass(frozen=True)
class ReviewStatus:
    """Per-problem status tracking data."""

    github_issue_id: int | None
    statement_reviewed_by: list[str]
    validators_reviewed_by: list[str]
    data_reviewed_by: list[str]


DEFAULT_REVIEW_STATUS = ReviewStatus(None, [], [], [])


DEFAULT_REVIEW_STATUS_TOML = """
# The GitHub Issue id for the problem, if there is one.
# github_issue_id = 

[review_status]
# Add your name in double quotes on a new line in this list if you have read
# the problem statement and believe that no further changes are necessary to
# the problem statement.
statement_reviewed_by = [

]
# Add your name in double quotes on a new line in this list if you have 
# reviewed the input validators and compared them to the problem statement and
# you believe that no further changes are necessary and all guarantees in the
# problem statement about the input are verified by an input validator.
validators_reviewed_by = [

]
# Add your name in double quotes on a new line in this list if you have 
# reviewed the test data and you believe that no further data is necessary to 
# ensure that incorrect solutions will be judged as WRONG ANSWER, too slow 
# solutions will be judged as TIME_LIMIT_EXCEEDED, and correct solutions will
# be judged as ACCEPTED.
data_reviewed_by = [

]
"""



"""Tests for the example problemsets."""

import os


def test_example_problemset(examples_path):
    """Test that the example problemset is being parsed as expected."""
    example_problemset_path = os.path.join(examples_path, "example_problemset")
    example_test_output = os.path.join(example_problemset_path, ".test_output")
    status = os.system(
        f"crifx {example_problemset_path} --output-dir {example_test_output}"
    )
    assert status == 0



"""Entry point for crifx."""

import argparse
import logging
import os
import sys

from crifx.config_parser import parse_config
from crifx.dir_layout_parsing import find_contest_problems_root
from crifx.git_manager import GitManager
from crifx.problemset_parser import ProblemSetParser
from crifx.report_writer import ReportWriter, make_crifx_dir

CRIFX_ERROR_EXIT_CODE = 1


def _dir_path_argparse_type(path):
    """Check that the provided path is a directory."""
    if os.path.isdir(path):
        return path
    else:
        raise ValueError(f"{path} is not a directory.")


def _make_argument_parser() -> argparse.ArgumentParser:
    """Create an argument parser."""
    parser = argparse.ArgumentParser(
        description="ICPC Contest preparation Reporting and Insights tool For anyone.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "path",
        type=_dir_path_argparse_type,
        nargs="?",
        default=None,
        help="Optional path to a problemset root directory. If not specified then "
        "crifx will test the current directory and up to 5 parent directories "
        "to find the first candidate problemset root directory.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Set verbose logging mode.",
    )
    parser.add_argument(
        "-o",
        "--output-dir",
        default=None,
        help="Optional directory path to which to write the crifx report pdf. "
        "If omitted, then the report will be written to the problemset "
        "root directory.",
    )
    return parser


def main():
    """Entry point for crifx."""
    args = _make_argument_parser().parse_args()
    log_level = logging.INFO
    if args.verbose:
        log_level = logging.DEBUG
    log_format = (
        "[%(asctime)s][%(levelname)s][%(name)s] %(message)s\t[%(filename)s:%(lineno)d]"
    )
    logging.basicConfig(level=log_level, format=log_format)
    logging.debug("Running crifx-cli from %s", os.getcwd())
    if args.path is None:
        problemset_root_path = find_contest_problems_root()
    else:
        problemset_root_path = os.path.abspath(args.path)
    if problemset_root_path is None:
        logging.error(
            "Could not find contest problems root from the current directory: %s",
            os.getcwd(),
        )
        sys.exit(CRIFX_ERROR_EXIT_CODE)
    if args.output_dir is None:
        output_dir = problemset_root_path
    else:
        output_dir = os.path.abspath(args.output_dir)
        if not os.path.isdir(output_dir):
            logging.error("Specified output directory '%s' does not exist", output_dir)
            sys.exit(CRIFX_ERROR_EXIT_CODE)
    config = parse_config(problemset_root_path)
    git_manager = GitManager(problemset_root_path)
    track_review_status = config.track_review_status
    problemset_parser = ProblemSetParser(
        problemset_root_path,
        git_manager,
        config.alias_groups,
        track_review_status,
    )
    problemset = problemset_parser.parse_problemset()
    writer = ReportWriter(problemset, config, git_manager)
    crifx_dir_path = make_crifx_dir(output_dir)
    writer.build_report(crifx_dir_path)
    writer.write_tex(crifx_dir_path)
    writer.write_pdf(output_dir)


if __name__ == "__main__":
    main()



"""Logic for interacting with git."""

import os
from collections import defaultdict
from dataclasses import dataclass

from pygit2 import Repository, Signature, discover_repository
from pygit2.enums import BlameFlag, FileStatus, SortMode


@dataclass(frozen=True)
class GitUser:
    """Data structure for git user information."""

    name: str
    email: str
    raw_name: bytes
    raw_email: bytes

    @staticmethod
    def from_signature(signature: Signature) -> "GitUser":
        """Create a GitUser object from a pygit2 Signature."""
        return GitUser(
            signature.name, signature.email, signature.raw_name, signature.raw_email
        )

    def __str__(self):
        """Get a string representation of a git user."""
        return f"{self.name} <{self.email}>"


class GitManager:
    """Manager class for interacting with git."""

    def __init__(self, path_in_repo: str):
        repo_path = discover_repository(path_in_repo)
        if repo_path is None:
            raise ValueError(f"Path '{path_in_repo}' is not in a git repository.")
        self.repo = Repository(repo_path)
        self.repo_root = os.path.abspath(os.path.join(repo_path, os.pardir))

    def get_committers_and_authors(self) -> list[GitUser]:
        """Get a list of every git user that has authored or committed a commit."""
        last_commit = self.repo[self.repo.head.target]  # type: ignore
        git_users = set()
        for commit in self.repo.walk(last_commit.id, SortMode.TIME):
            author = commit.author
            committer = commit.committer
            git_users.add(GitUser.from_signature(author))
            git_users.add(GitUser.from_signature(committer))
        return sorted(git_users, key=lambda x: x.name)

    def guess_file_author(self, abs_path: str) -> GitUser | None:
        """Guess the author of a file path."""
        if not os.path.isfile(abs_path):
            raise ValueError(f"Path '{abs_path}' is not a file.")
        path = os.path.relpath(abs_path, self.repo_root)
        file_status = self.repo.status_file(path)
        if file_status in (FileStatus.WT_NEW, FileStatus.INDEX_NEW):
            # Handle cases where the file is new and untracked or staged but
            # not committed. Assume that the current git user is the author.
            name = self.repo.config.get_global_config()["user.name"]  # type: ignore
            email = self.repo.config.get_global_config()["user.email"]  # type: ignore
            if name is not None and email is not None:
                return GitUser.from_signature(Signature(name, email))
            return None
        blame = self.repo.blame(  # type: ignore
            path, flags=BlameFlag.NORMAL | BlameFlag.IGNORE_WHITESPACE
        )
        lines_modified: defaultdict[GitUser, int] = defaultdict(int)
        lines_max = 0
        user_max = None
        for hunk in blame:
            committer = hunk.final_committer
            git_user = GitUser.from_signature(committer)
            lines_modified[git_user] += hunk.lines_in_hunk
            if lines_modified[git_user] > lines_max:
                lines_max = lines_modified[git_user]
                user_max = git_user
        return user_max

    def get_commit_id(self):
        """Get the current commit id."""
        return self.repo.head.target

    def get_short_commit_id(self):
        """Get the first 8 characters of the current commit id."""
        commid_id_str = str(self.get_commit_id())
        return commid_id_str[:8]



"""Tests for writing the report."""

import os

from crifx.config_parser import Config
from crifx.contest_objects import ProblemSet
from crifx.git_manager import GitManager
from crifx.report_writer import ReportWriter


def test_write_report(tmp_path, scenarios_path):
    """A report tex file can be written to the expected path."""
    assert not os.listdir(tmp_path)
    problemset = ProblemSet([])
    git_manager = GitManager(scenarios_path)
    config = Config({})
    writer = ReportWriter(problemset, config, git_manager)
    writer.build_report(tmp_path)
    writer.write_tex(tmp_path)
    expected_file = "crifx-report.tex"
    assert expected_file in os.listdir(tmp_path)
    report_path = os.path.join(tmp_path, expected_file)
    with open(report_path, "r") as tmp_file:
        lines = tmp_file.readlines()
    expected_title = "\\title{CRIFX Contest Preparation Status Report}%\n"
    assert expected_title in lines



"""Objects corresponding to entities in a contest problemset package."""

from crifx.contest_objects.judge import UNKNOWN_JUDGE, Judge
from crifx.contest_objects.judgement import Judgement
from crifx.contest_objects.problem import Problem
from crifx.contest_objects.problem_test_case import ProblemTestCase
from crifx.contest_objects.problemset import ProblemSet
from crifx.contest_objects.programming_language import (
    LanguageGroup,
    ProgrammingLanguage,
)
from crifx.contest_objects.submission import Submission

__all__ = [
    "Judge",
    "Judgement",
    "LanguageGroup",
    "Problem",
    "ProblemSet",
    "ProblemTestCase",
    "ProgrammingLanguage",
    "Submission",
    "UNKNOWN_JUDGE",
]



"""Data structure and manager for a problem set."""

from crifx.contest_objects.judge import Judge
from crifx.contest_objects.problem import Problem


class ProblemSet:
    """Data structure and manager for a problem set."""

    def __init__(self, problems: list[Problem]):
        sorted_problems = sorted(problems, key=lambda x: x.name)
        self.problems = sorted_problems

    def submission_authors(self) -> list[Judge]:
        """Get the set of authors that have contributed at least one submission."""
        authors = set()
        for problem in self.problems:
            for submission in problem.submissions:
                authors.add(submission.author)
        return sorted(authors, key=lambda x: x.primary_name)



"""Data structure for data about a problem."""

from collections import defaultdict
from dataclasses import dataclass

from crifx.contest_objects.judge import Judge
from crifx.contest_objects.judgement import Judgement
from crifx.contest_objects.problem_test_case import ProblemTestCase
from crifx.contest_objects.programming_language import (
    LanguageGroup,
    ProgrammingLanguage,
)
from crifx.contest_objects.submission import Submission
from crifx.report_objects import ReviewStatus


@dataclass(frozen=True)
class Problem:
    """Data for a Problem."""

    name: str
    test_cases: list[ProblemTestCase]
    submissions: list[Submission]
    review_status: ReviewStatus

    def _get_submissions_with_judgement(self, judgement):
        return list(filter(lambda x: x.judgement is judgement, self.submissions))

    @property
    def ac_submissions(self) -> list[Submission]:
        """Get the AC submissions."""
        return self._get_submissions_with_judgement(Judgement.ACCEPTED)

    @property
    def wa_submissions(self) -> list[Submission]:
        """Get the WA submissions."""
        return self._get_submissions_with_judgement(Judgement.WRONG_ANSWER)

    @property
    def tle_submissions(self) -> list[Submission]:
        """Get the TLE submissions."""
        return self._get_submissions_with_judgement(Judgement.TIME_LIMIT_EXCEEDED)

    @property
    def rte_submissions(self) -> list[Submission]:
        """Get the RTE submissions."""
        return self._get_submissions_with_judgement(Judgement.RUN_TIME_ERROR)

    def independent_ac_count(self) -> int:
        """Get the number of AC submissions by different authors."""
        accepted_authors: list[Judge] = []
        for submission in self.ac_submissions:
            if any(submission.author.is_same(author) for author in accepted_authors):
                continue
            accepted_authors.append(submission.author)
        return len(accepted_authors)

    def ac_lines_of_code_min(self):
        """Get the minimum Lines of Code among AC submissions to this problem."""
        ac_min = None
        for submission in self.ac_submissions:
            if ac_min is None or submission.lines_of_code < ac_min:
                ac_min = submission.lines_of_code
        return ac_min

    def ac_lines_of_code_median(self):
        """Get the median Lines of Code among AC submissions to this problem."""
        ac_loc_vals = []
        for submission in self.ac_submissions:
            ac_loc_vals.append(submission.lines_of_code)
        ac_loc_vals.sort()
        middle = len(ac_loc_vals) // 2
        if len(ac_loc_vals) % 2 == 0:
            return (ac_loc_vals[middle - 1] + ac_loc_vals[middle]) // 2
        return ac_loc_vals[middle]

    def ac_languages(self) -> defaultdict[ProgrammingLanguage, int]:
        """Get the number of AC submissions in each language."""
        language_counts: defaultdict[ProgrammingLanguage, int] = defaultdict(int)
        for submission in self.ac_submissions:
            language_counts[submission.language] += 1
        return language_counts

    def language_groups_ac_covered(self, language_groups: list[LanguageGroup]):
        """Get the number of language groups that have at least one AC submission."""
        groups_covered = []
        ac_languages = self.ac_languages()
        for language_group in language_groups:
            covered = False
            for language, count in ac_languages.items():
                if count > 0 and language_group.has_language(language):
                    covered = True
                    break
            if covered:
                groups_covered.append(language_group)
        return groups_covered



"""Tests for file hierarchy parsing."""

import os
import unittest.mock as mock
import uuid

from crifx.dir_layout_parsing import (
    find_contest_problems_root,
    get_problem_root_dirs,
    is_contest_problems_root,
    is_problem_root_dir,
)


def test_is_problem_root_dir(tmp_path):
    """Test detecting a problem directory."""
    # A directory with a submissions directory inside is a problem root dir.
    dir_1 = os.path.join(tmp_path, str(uuid.uuid4()))
    os.mkdir(dir_1)
    os.mkdir(os.path.join(dir_1, "submissions"))
    assert is_problem_root_dir(dir_1)

    # A directory with a problem_statement directory inside is a problem root dir.
    dir_2 = os.path.join(tmp_path, str(uuid.uuid4()))
    os.mkdir(dir_2)
    os.mkdir(os.path.join(dir_2, "problem_statement"))
    assert is_problem_root_dir(dir_2)

    # A directory with a problem.yaml file is a problem root dir.
    dir_3 = os.path.join(tmp_path, str(uuid.uuid4()))
    os.mkdir(dir_3)
    open(os.path.join(dir_3, "problem.yaml"), "a").close()
    assert is_problem_root_dir(dir_3)

    # A directory with a problem.yml file is a problem root dir.
    dir_4 = os.path.join(tmp_path, str(uuid.uuid4()))
    os.mkdir(dir_4)
    open(os.path.join(dir_4, "problem.yml"), "a").close()
    assert is_problem_root_dir(dir_4)

    # Dir with no submissions, problem_statement, nor problem yaml is not problem root.
    dir_5 = os.path.join(tmp_path, str(uuid.uuid4()))
    os.mkdir(dir_5)
    os.mkdir(os.path.join(dir_5, "data"))
    assert not is_problem_root_dir(dir_5)

    # Non-dir file is not a problem root dir.
    tmp_file = os.path.join(tmp_path, f"{str(uuid.uuid4())}.txt")
    open(tmp_file, "a").close()
    assert not is_problem_root_dir(tmp_file)


def test_get_problem_root_dirs(tmp_path, make_problem_skeleton_dir):
    """Contest problem paths can be identified in a contest problems directory."""
    assert not os.listdir(tmp_path)
    # Set up a couple of problems.
    path_1 = make_problem_skeleton_dir()
    path_2 = make_problem_skeleton_dir()
    # Problem skeletons were set up in the tmp_path.
    assert os.listdir(tmp_path)
    # Add some other arbitrary files.
    open(os.path.join(tmp_path, "problemset.pdf"), "a").close()
    open(os.path.join(tmp_path, "random_file.txt"), "a").close()
    assert is_contest_problems_root(tmp_path)
    assert sorted(get_problem_root_dirs(tmp_path)) == sorted([path_1, path_2])


def test_find_contest_problem_root(tmp_path, make_problem_skeleton_dir):
    """The contest problem root can be found from inside nested directories."""
    assert not os.listdir(tmp_path)
    problem_path = make_problem_skeleton_dir()
    assert os.listdir(tmp_path)
    with mock.patch(
        "os.getcwd", return_value=os.path.join(problem_path, "submissions/accepted")
    ):
        assert find_contest_problems_root() == str(tmp_path)
    with mock.patch(
        "os.getcwd", return_value=os.path.join(problem_path, "submissions")
    ):
        assert find_contest_problems_root() == str(tmp_path)
    with mock.patch(
        "os.getcwd", return_value=os.path.join(problem_path, "submissions/data/secret")
    ):
        assert find_contest_problems_root() == str(tmp_path)



"""Object for representing a submission author."""

from crifx.git_manager import GitUser


class Judge:
    """Object for representing a judge submission author."""

    def __init__(self, primary_name: str, git_user: GitUser | None, *aliases: str):
        self.primary_name: str = primary_name
        self.git_name: str | None = getattr(git_user, "name", None)
        name_set = {name for name in aliases}
        name_set.add(primary_name)
        if self.git_name is not None:
            name_set.add(self.git_name)
        self.aliases = [name.lower() for name in name_set]

    def has_alias(self, name: str):
        """Determine if the provided name corresponds to this judge."""
        return name.lower() in self.aliases

    def is_same(self, other_judge):
        """Determine if two Judge objects correspond to the same person."""
        return (
            self.git_name == other_judge.git_name
            or self.primary_name == other_judge.primary_name
        )

    def __str__(self):
        return f"{self.primary_name}"


UNKNOWN_JUDGE = Judge("UNKNOWN", None)



"""Module for writing the report to file."""

import logging
import os

from pylatex import (
    Command,
    Document,
    Enumerate,
    Itemize,
    MultiColumn,
    NoEscape,
    Section,
    Subsection,
    Subsubsection,
    Tabular,
)
from pylatex.base_classes import Environment
from pylatex.package import Package

from crifx.config_parser import Config
from crifx.contest_objects import Problem, ProblemSet
from crifx.git_manager import GitManager

MARGIN = "2cm"
REPORT_FILENAME = "crifx-report"
INPUT_FILE_LINES_MAX = 10
INPUT_FILE_WIDTH_MAX = 90

LISTING_OPTIONS = [
    NoEscape(r"basicstyle=\footnotesize"),
    NoEscape(r"backgroundcolor=\color{lightgray}"),
    "framexleftmargin=1em",
    "framexrightmargin=1em",
    "breaklines=true",
]


def truncate(word, characters_max=10):
    """Get a truncated string representation."""
    if len(word) > characters_max:
        return word[: characters_max - 2] + "..."
    return word


class LstListing(Environment):
    """LstListing environment."""

    packages = [Package("listings")]
    escape = False
    content_separator = "\n"


class ReportWriter:
    """Manager class for writing the crifx report."""

    def __init__(
        self, problem_set: ProblemSet, config: Config, git_manager: GitManager
    ):
        self.problem_set = problem_set
        self.crifx_config = config
        self.git_manager = git_manager
        self.doc: Document | None = None

    def build_report(self, crifx_dir_path: str) -> Document:
        """Build the report."""
        report_tex_path = os.path.join(crifx_dir_path, REPORT_FILENAME)
        geometry_options = {
            "tmargin": MARGIN,
            "lmargin": MARGIN,
            "rmargin": MARGIN,
            "bmargin": MARGIN,
        }
        self.doc = Document(report_tex_path, geometry_options=geometry_options)
        self._set_preamble()
        self._write_body()
        return self.doc

    def _set_preamble(self):
        """Set the preamble for the tex document."""
        git_short_commit_id = self.git_manager.get_short_commit_id()
        self.doc.preamble.append(Command("usepackage", "datetime2"))
        self.doc.preamble.append(Command("usepackage", "listings"))
        self.doc.preamble.append(
            Command(
                "usepackage",
                ("hyperref",),
                ("colorlinks=true", "urlcolor=blue", "linkcolor=blue"),
            )
        )
        self.doc.preamble.append(Command("usepackage", "fancyhdr"))
        self.doc.preamble.append(Command("pagestyle", ("fancy",)))
        self.doc.preamble.append(Command("label", "TOP"))
        self.doc.preamble.append(
            Command(
                "fancyhead",
                (Command("hyperref", ("Back to Top",), ("TOP",)),),
                ("HR",),
            )
        )
        self.doc.preamble.append(
            Command("definecolor", ("insufficientred", "RGB", "255,100,100"))
        )
        self.doc.preamble.append(
            Command("definecolor", ("sufficientgreen", "RGB", "0,210,0"))
        )
        self.doc.preamble.append(
            Command("title", "CRIFX Contest Preparation Status Report")
        )
        self.doc.preamble.append(Command("author", ""))
        self.doc.preamble.append(
            Command(
                "date",
                NoEscape(
                    f"Compiled \\today~at \\DTMcurrenttime\\DTMcurrentzone~"
                    f"for commit {git_short_commit_id}"
                ),
            )
        )

    def _write_body(self):
        """Write the body of the document."""
        self.doc.append(NoEscape(r"\maketitle"))
        self._write_summary_table()
        self._write_manual_reviews_table()
        self._write_how_can_i_help()
        for problem in self.problem_set.problems:
            self._write_problem_details(problem)

    def _write_summary_table(self):
        """Write the summary table for the document."""
        language_group_configs = self.crifx_config.language_group_configs
        language_groups = [
            group_config.language_group for group_config in language_group_configs
        ]
        requirements = self.crifx_config.review_requirements
        num_columns = 9 + len(language_group_configs)
        column_spec = "|l|" + "c|" * (num_columns - 1)
        with self.doc.create(Section("Submissions summary", numbering=False)):
            with self.doc.create(Tabular(column_spec)) as table:
                table.add_hline()
                header_group_row = [
                    # Problem
                    "",
                    # Independent, Groups, language_groups, Sum
                    MultiColumn(
                        3 + len(language_group_configs),
                        align="c|",
                        data=NoEscape(r"{\tiny Solutions}"),
                    ),
                    # WA, TLE
                    MultiColumn(2, align="c|", data=NoEscape(r"{\tiny Non-solutions}")),
                    # AC LOC.
                    MultiColumn(
                        2, align="c|", data=NoEscape(r"{\tiny AC Lines of Code}")
                    ),
                    # Test cases
                    "",
                ]
                table.add_row(header_group_row, color="cyan")
                table.add_hline()
                header_row = [
                    NoEscape(r"{\tiny Problem}"),
                    NoEscape(r"{\tiny Independent}"),
                    NoEscape(r"{\tiny Lang. Groups}"),
                ]
                for language_group_config in language_group_configs:
                    header_row.append(
                        NoEscape(
                            r"{\tiny "
                            + truncate(language_group_config.identifier, 10)
                            + r"}"
                        )
                    )
                header_row.extend(
                    [
                        NoEscape(r"{\tiny Sum}"),
                        NoEscape(r"{\tiny WA}"),
                        NoEscape(r"{\tiny TLE}"),
                        NoEscape(r"{\tiny Min.}"),
                        NoEscape(r"{\tiny Med.}"),
                        NoEscape(r"{\tiny Test Files}"),
                    ]
                )
                table.add_row(
                    header_row,
                    color="cyan",
                )
                table.add_hline()
                for problem in self.problem_set.problems:
                    row = [
                        Command(
                            "hyperref",
                            (truncate(problem.name, 16),),
                            (f"sec:{problem.name}",),
                        ),
                        self._coloured_cell(
                            problem.independent_ac_count(), requirements.independent_ac
                        ),
                        self._coloured_cell(
                            len(problem.language_groups_ac_covered(language_groups)),
                            requirements.language_groups_ac,
                        ),
                    ]
                    for language_group_config in language_group_configs:
                        language_group = language_group_config.language_group
                        count = 0
                        for ac_submission in problem.ac_submissions:
                            if language_group.has_language(ac_submission.language):
                                count += 1
                        row.append(
                            self._coloured_cell(
                                count, language_group_config.required_ac_count
                            )
                        )
                    row.extend(
                        [
                            len(problem.ac_submissions),
                            len(problem.wa_submissions),
                            len(problem.tle_submissions),
                            str(problem.ac_lines_of_code_min()),
                            str(problem.ac_lines_of_code_median()),
                            len(problem.test_cases),
                        ]
                    )
                    table.add_row(row)
                    table.add_hline()

    def _write_manual_reviews_table(self):
        """Write a table with a summary tracking manual reviews."""
        requirements = self.crifx_config.review_requirements
        show_statement_reviews = requirements.statement_reviewers > 0
        show_validator_reviews = requirements.validator_reviewers > 0
        show_data_reviews = requirements.data_reviewers > 0
        review_columns = (
            int(show_statement_reviews)
            + int(show_validator_reviews)
            + int(show_data_reviews)
        )
        if review_columns == 0:
            return
        num_columns = 1 + review_columns
        column_spec = "|l|" + "c|" * (num_columns - 1)
        with self.doc.create(Section("Manual review tracking", numbering=False)):
            with self.doc.create(Tabular(column_spec)) as table:
                table.add_hline()
                header_row = [NoEscape(r"{\tiny Problem}")]
                if show_statement_reviews:
                    header_row.append(NoEscape(r"{\tiny Statement}"))
                if show_validator_reviews:
                    header_row.append(NoEscape(r"{\tiny Validator(s)}"))
                if show_data_reviews:
                    header_row.append(NoEscape(r"{\tiny Data}"))
                table.add_row(
                    header_row,
                    color="cyan",
                )
                table.add_hline()
                for problem in self.problem_set.problems:
                    row = [
                        Command("hyperref", (problem.name,), (f"sec:{problem.name}",))
                    ]
                    if show_statement_reviews:
                        row.append(
                            self._coloured_cell(
                                len(problem.review_status.statement_reviewed_by),
                                requirements.statement_reviewers,
                            )
                        )
                    if show_validator_reviews:
                        row.append(
                            self._coloured_cell(
                                len(problem.review_status.validators_reviewed_by),
                                requirements.validator_reviewers,
                            )
                        )
                    if show_data_reviews:
                        row.append(
                            self._coloured_cell(
                                len(problem.review_status.data_reviewed_by),
                                requirements.data_reviewers,
                            )
                        )
                    table.add_row(row)
                    table.add_hline()

    @staticmethod
    def _coloured_cell(value: int, requirement: int) -> int | str | NoEscape:
        if requirement == 0:
            return value
        if value < requirement:
            return NoEscape(r"\cellcolor{insufficientred}" + f"{value}/{requirement}")
        else:
            return NoEscape(r"\cellcolor{sufficientgreen}" + f"{value}/{requirement}")

    @staticmethod
    def _oxford_list(items: list[str], connector: str):
        """Get a text list using the oxford comma."""
        if len(items) == 1:
            return items[0]
        if len(items) == 2:
            return f"{items[0]} {connector} {items[1]}"
        joined_names = ", ".join(items[:-1])
        return f"{joined_names}, {connector} {items[-1]}"

    def _oxford_and(self, items: list[str]):
        """Get a text 'and' list using the oxford comma."""
        return self._oxford_list(items, "and")

    def _oxford_or(self, items: list[str]):
        """Get a text 'or' list using the oxford comma."""
        return self._oxford_list(items, "or")

    def _add_independent_ac_needs(self, enum_env, problem):
        """Add text list of independent AC submission needs."""
        requirements = self.crifx_config.review_requirements
        if problem.independent_ac_count() < requirements.independent_ac:
            independent_needed = (
                requirements.independent_ac - problem.independent_ac_count()
            )
            ac_judge_names = {
                submission.author.primary_name for submission in problem.ac_submissions
            }
            if not ac_judge_names:
                if independent_needed == 1:
                    enum_env.add_item(f"{problem.name} needs an AC submission.")
                else:
                    enum_env.add_item(
                        f"{problem.name} needs {independent_needed} AC " f"submissions."
                    )
            elif independent_needed == 1:
                if requirements.independent_ac == 1:
                    enum_env.add_item(f"{problem.name} needs an AC submission.")
                else:
                    enum_env.add_item(
                        f"{problem.name} needs at least one more AC submission "
                        f"from someone other than "
                        f"{self._oxford_and(sorted(ac_judge_names))}."
                    )
            else:
                enum_env.add_item(
                    f"{problem.name} needs at least {independent_needed} more "
                    f"AC submissions from people other than "
                    f"{self._oxford_and(sorted(ac_judge_names))}."
                )

    def _add_language_group_ac_needs(self, enum_env, problem):
        """Add text list of language group AC submission needs."""
        requirements = self.crifx_config.review_requirements
        language_group_configs = self.crifx_config.language_group_configs
        language_groups = [
            group_config.language_group for group_config in language_group_configs
        ]
        groups_covered = problem.language_groups_ac_covered(language_groups)
        if len(groups_covered) < requirements.language_groups_ac:
            groups_needed_num = requirements.language_groups_ac - len(groups_covered)
            groups_not_covered_names = [
                config.identifier
                for config in language_group_configs
                if config.language_group not in groups_covered
            ]
            if groups_needed_num == 1:
                enum_env.add_item(
                    f"{problem.name} needs at least one more AC submission from "
                    f"any of the following language groups: "
                    f"{self._oxford_or(groups_not_covered_names)}."
                )
            else:
                enum_env.add_item(
                    f"{problem.name} needs at least {groups_needed_num} more AC "
                    f"submissions from any of the following language groups: "
                    f"{self._oxford_or(groups_not_covered_names)}."
                )

    def _add_tle_needs(self, enum_env, problem):
        """Add a text line for TLE submission needs for a problem."""
        requirements = self.crifx_config.review_requirements
        if len(problem.tle_submissions) < requirements.submissions_tle:
            tle_needed = requirements.submissions_tle - len(problem.tle_submissions)
            if requirements.submissions_tle == 1:
                enum_env.add_item(f"{problem.name} needs at least one TLE submission.")
            elif tle_needed == 1:
                enum_env.add_item(
                    f"{problem.name} needs at least one more TLE submission."
                )
            else:
                enum_env.add_item(
                    f"{problem.name} needs at least {tle_needed} more TLE "
                    f"submissions."
                )

    def _add_wa_needs(self, enum_env, problem):
        """Add a text line for WA submission needs for each problem."""
        requirements = self.crifx_config.review_requirements
        if len(problem.wa_submissions) < requirements.submissions_wa:
            wa_needed = requirements.submissions_wa - len(problem.wa_submissions)
            if requirements.submissions_wa == 1:
                enum_env.add_item(f"{problem.name} needs at least one WA submission.")
            elif wa_needed == 1:
                enum_env.add_item(
                    f"{problem.name} needs at least one more WA submission."
                )
            else:
                enum_env.add_item(
                    f"{problem.name} needs at least {wa_needed} more WA "
                    f"submissions."
                )

    def _add_review_needs(
        self, enum_env, problem_name, reviewers, required_count, review_type
    ):
        """Add an item for review needs of a given type."""
        if len(reviewers) < required_count:
            reviews_needed = required_count - len(reviewers)
            if not reviewers:
                if required_count == 1:
                    enum_env.add_item(
                        f"{problem_name} needs at least one {review_type} review."
                    )
                else:
                    enum_env.add_item(
                        f"{problem_name} needs at least {reviews_needed} "
                        f"{review_type} reviews."
                    )
            elif required_count == 1:
                enum_env.add_item(
                    f"{problem_name} needs at least one {review_type} review."
                )
            elif reviews_needed == 1:
                enum_env.add_item(
                    f"{problem_name} needs at least one more {review_type} review "
                    f"from someone other than {self._oxford_and(reviewers)}."
                )
            else:
                enum_env.add_item(
                    f"{problem_name} needs at least {reviews_needed} more "
                    f"{review_type} reviews from people other than "
                    f"{self._oxford_and(reviewers)}."
                )

    def _add_statement_review_needs(self, enum_env, problem):
        """Add an item for statement reviews needed for a problem."""
        requirements = self.crifx_config.review_requirements
        required_count = requirements.statement_reviewers
        reviewers = problem.review_status.statement_reviewed_by
        self._add_review_needs(
            enum_env, problem.name, reviewers, required_count, "statement"
        )

    def _add_data_review_needs(self, enum_env, problem):
        """Add an item for test data reviews needed for a problem."""
        requirements = self.crifx_config.review_requirements
        required_count = requirements.data_reviewers
        reviewers = problem.review_status.data_reviewed_by
        self._add_review_needs(
            enum_env, problem.name, reviewers, required_count, "test data"
        )

    def _add_validator_review_needs(self, enum_env, problem):
        """Add an item for validator reviews needed for each problem."""
        requirements = self.crifx_config.review_requirements
        required_count = requirements.validator_reviewers
        reviewers = problem.review_status.validators_reviewed_by
        self._add_review_needs(
            enum_env, problem.name, reviewers, required_count, "validator"
        )

    def _write_how_can_i_help(self):
        """Write the 'How can I help?' section."""
        with self.doc.create(Section("How can I help?", numbering=False)):
            with self.doc.create(Enumerate()) as enum_env:
                for problem in self.problem_set.problems:
                    self._add_independent_ac_needs(enum_env, problem)
                for problem in self.problem_set.problems:
                    self._add_language_group_ac_needs(enum_env, problem)
                for problem in self.problem_set.problems:
                    self._add_tle_needs(enum_env, problem)
                for problem in self.problem_set.problems:
                    self._add_wa_needs(enum_env, problem)
                for problem in self.problem_set.problems:
                    self._add_statement_review_needs(enum_env, problem)
                for problem in self.problem_set.problems:
                    self._add_validator_review_needs(enum_env, problem)
                for problem in self.problem_set.problems:
                    self._add_data_review_needs(enum_env, problem)
                enum_env.add_item("Add test data")
                enum_env.add_item("Add input validators")

    def _write_problem_details(self, problem: Problem):
        """Write the details for a problem."""
        assert self.doc is not None
        self.doc.append(Command(r"newpage"))
        with self.doc.create(Section(problem.name)):
            with self.doc.create(
                Subsection("How can I help?", numbering=False, label=False)
            ):
                with self.doc.create(Enumerate()) as enum_env:
                    self._add_independent_ac_needs(enum_env, problem)
                    self._add_language_group_ac_needs(enum_env, problem)
                    self._add_tle_needs(enum_env, problem)
                    self._add_wa_needs(enum_env, problem)
                    self._add_statement_review_needs(enum_env, problem)
                    self._add_validator_review_needs(enum_env, problem)
                    self._add_data_review_needs(enum_env, problem)
            with self.doc.create(
                Subsection("Submissions", numbering=False, label=False)
            ):
                with self.doc.create(
                    Subsubsection("Accepted", numbering=False, label=False)
                ):
                    if not problem.ac_submissions:
                        self.doc.append("No accepted submissions.")
                    with self.doc.create(Itemize()) as itemize:
                        for submission in problem.ac_submissions:
                            itemize.add_item(
                                f"{submission.filename} by {submission.author}. "
                                f"{submission.lines_of_code} lines of code."
                            )
                with self.doc.create(
                    Subsubsection("Wrong Answer", numbering=False, label=False)
                ):
                    if not problem.wa_submissions:
                        self.doc.append("No wrong answer submissions.")
                    with self.doc.create(Itemize()) as itemize:
                        for submission in problem.wa_submissions:
                            itemize.add_item(
                                f"{submission.filename} by {submission.author}. "
                                f"{submission.lines_of_code} lines of code."
                            )
                with self.doc.create(
                    Subsubsection("Time Limit Exceeded", numbering=False, label=False)
                ):
                    if not problem.tle_submissions:
                        self.doc.append("No time limit exceeded submissions.")
                    with self.doc.create(Itemize()) as itemize:
                        for submission in problem.tle_submissions:
                            itemize.add_item(
                                f"{submission.filename} by {submission.author}. "
                                f"{submission.lines_of_code} lines of code."
                            )
            with self.doc.create(
                Subsection("Test Cases", numbering=False, label=False)
            ):
                self.doc.append(
                    "Test case descriptions are rendered below if .desc files exist."
                )
                with self.doc.create(Itemize()) as itemize:
                    for test_case in problem.test_cases:
                        itemize.add_item(test_case.name)
                        if test_case.has_description:
                            desc_filepath = os.path.join(
                                test_case.dir_path, f"{test_case.name}.desc"
                            )
                            itemize.append(
                                Command(
                                    "lstinputlisting",
                                    NoEscape(desc_filepath),
                                    options=LISTING_OPTIONS,
                                )
                            )

    def write_tex(self, dirpath: str):
        """Write the tex output."""
        if self.doc is None:
            raise ValueError(
                "The tex file cannot be written yet. The document has not been built."
            )
        filepath = os.path.join(dirpath, REPORT_FILENAME)
        logging.debug("Writing tex to %s", filepath)
        self.doc.generate_tex(filepath)

    def write_pdf(self, dirpath: str):
        """Write a pdf file from the tex file."""
        if self.doc is None:
            raise ValueError(
                "The pdf file cannot be written yet. The document has not been built."
            )
        filepath = os.path.join(dirpath, REPORT_FILENAME)
        logging.debug("Writing pdf to %s", filepath)
        self.doc.generate_pdf(filepath, clean=True, clean_tex=True)


def make_crifx_dir(containing_dir_path: str) -> str:
    """Create the crifx directory."""
    crifx_dir_path = os.path.join(containing_dir_path, ".crifx")
    if not os.path.exists(crifx_dir_path):
        os.mkdir(crifx_dir_path)
    return crifx_dir_path



"""Tests for parsing problemsets."""

import os

from crifx.config_parser import parse_config
from crifx.contest_objects import ProgrammingLanguage
from crifx.git_manager import GitManager
from crifx.problemset_parser import ProblemSetParser


def test_scenario_1(scenarios_path):
    """Test the basic scenario 1."""
    path = os.path.join(scenarios_path, "sample_contest_1")
    git_manager = GitManager(scenarios_path)
    config = parse_config(path)
    parser = ProblemSetParser(path, git_manager, config.alias_groups, False)
    problemset = parser.parse_problemset()
    problem_a = next(
        problem for problem in problemset.problems if problem.name == "problem_a"
    )
    assert len(problemset.problems) == 2
    assert len(problem_a.test_cases) == 5
    a_desc_test_cases = list(filter(lambda x: x.has_description, problem_a.test_cases))
    assert len(a_desc_test_cases) == 1
    desc_test_case = a_desc_test_cases[0]
    assert desc_test_case.name == "2"
    assert desc_test_case.dir_path.endswith("data/secret")
    assert desc_test_case.description_lines == [
        "Secret test case\n",
        "\n",
        "this is a description with a blank line\n",
    ]
    assert desc_test_case.image_extension is None
    assert len(problem_a.submissions) == 3
    assert len(problem_a.ac_submissions) == 2
    assert len(problem_a.wa_submissions) == 1
    assert len(problem_a.tle_submissions) == 0

    problem_a_java_ac = next(
        sub
        for sub in problem_a.ac_submissions
        if sub.language is ProgrammingLanguage.JAVA
    )
    assert problem_a_java_ac.author.primary_name == "Jane Doe"



"""Module for detecting and parsing the problem and contest directory structure."""

import os

PROBLEM_ROOT_INDICATOR_DIRS = [
    "submissions",
    "problem_statement",
]

PROBLEM_ROOT_INDICATOR_FILES = [
    "problem.yaml",
    "problem.yml",
]


def is_problem_root_dir(path: str) -> bool:
    """Detect if the given path is the root of a problem directory."""
    if not os.path.isdir(path):
        return False
    for dir_obj_name in os.listdir(path):
        dir_obj_path = os.path.join(path, dir_obj_name)
        if os.path.isdir(dir_obj_path) and dir_obj_name in PROBLEM_ROOT_INDICATOR_DIRS:
            return True
        if (
            os.path.isfile(dir_obj_path)
            and dir_obj_name in PROBLEM_ROOT_INDICATOR_FILES
        ):
            return True
    return False


def get_problem_root_dirs(path: str) -> list[str]:
    """Get the problem root directory paths under the current directory."""
    if not os.path.isdir(path):
        return []
    problem_root_dirs = []
    try:
        for dir_obj_name in os.listdir(path):
            dir_obj_path = os.path.join(path, dir_obj_name)
            if os.path.isdir(dir_obj_path) and is_problem_root_dir(dir_obj_path):
                problem_root_dirs.append(dir_obj_path)
        return problem_root_dirs
    except PermissionError:
        return []


def is_contest_problems_root(path: str) -> bool:
    """Detect if a path has one or more problem root directories."""
    return bool(get_problem_root_dirs(path))


def find_contest_problems_root() -> str | None:
    """
    Find the contest problems directory path from the current working directory.

    Return `None` if the directory is not found within 5 parent levels.
    """
    current_dir = os.getcwd()
    candidate_dir = current_dir
    parents_max = 5
    for _ in range(parents_max):
        if is_contest_problems_root(candidate_dir):
            return candidate_dir
        candidate_dir = os.path.dirname(candidate_dir)
    return None



"""Tests for the GitManager class."""

import os

from crifx.git_manager import GitManager


def test_guess_existing_file_author():
    """The guess_file_author method returns a git user."""
    path = os.path.join(
        "tests",
        "scenarios",
        "sample_contest_1",
        "problem_a",
        "submissions",
        "accepted",
        "hello_world.py",
    )
    git_manager = GitManager(os.getcwd())

    git_user = git_manager.guess_file_author(path)
    assert git_user.name == "Finn Lidbetter"


def test_guess_untracked_file_author(
    make_file_at_path, empty_repo, global_git_config_path
):
    """Guessing the file author of an untracked file works."""
    with open(global_git_config_path, "w") as global_git_config:
        global_git_config.writelines(
            [
                "[user]\n",
                "name = Test User\n",
                "email = tester@example.com\n",
                "[init]\n",
                "defaultBranch = main\n",
            ]
        )

    repo_git_path = empty_repo.path
    git_manager = GitManager(repo_git_path)

    written_file = make_file_at_path(git_manager.repo_root)

    git_user = git_manager.guess_file_author(written_file)
    assert git_user is not None
    assert git_user.name == "Test User"
    assert git_user.email == "tester@example.com"



"""Submission judgement enumeration object."""

from enum import Enum


class Judgement(Enum):
    """Classification of a submission."""

    ACCEPTED = "accepted"
    TIME_LIMIT_EXCEEDED = "time_limit_exceeded"
    RUN_TIME_ERROR = "run_time_error"
    WRONG_ANSWER = "wrong_answer"

    def abbreviation(self) -> str:
        """Get a short form name for the judgement."""
        match self:
            case Judgement.ACCEPTED:
                return "AC"
            case Judgement.TIME_LIMIT_EXCEEDED:
                return "TLE"
            case Judgement.RUN_TIME_ERROR:
                return "RTE"
            case Judgement.WRONG_ANSWER:
                return "WA"
